//**************************************************************************
//**
//**	##   ##    ##    ##   ##   ####     ####   ###     ###
//**	##   ##  ##  ##  ##   ##  ##  ##   ##  ##  ####   ####
//**	 ## ##  ##    ##  ## ##  ##    ## ##    ## ## ## ## ##
//**	 ## ##  ########  ## ##  ##    ## ##    ## ##  ###  ##
//**	  ###   ##    ##   ###    ##  ##   ##  ##  ##       ##
//**	   #    ##    ##    #      ####     ####   ##       ##
//**
//**	$Id: cl_parse.cpp 1655 2006-08-12 14:05:30Z dj_jl $
//**
//**	Copyright (C) 1999-2006 Jānis Legzdiņš
//**
//**	This program is free software; you can redistribute it and/or
//**  modify it under the terms of the GNU General Public License
//**  as published by the Free Software Foundation; either version 2
//**  of the License, or (at your option) any later version.
//**
//**	This program is distributed in the hope that it will be useful,
//**  but WITHOUT ANY WARRANTY; without even the implied warranty of
//**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//**  GNU General Public License for more details.
//**
//**************************************************************************

// HEADER FILES ------------------------------------------------------------

#include "gamedefs.h"
#include "cl_local.h"

// MACROS ------------------------------------------------------------------

// TYPES -------------------------------------------------------------------

// EXTERNAL FUNCTION PROTOTYPES --------------------------------------------

void CL_ClearInput();
void CL_PO_SpawnPolyobj(float x, float y, int tag);
void CL_PO_TranslateToStartSpot(float originX, float originY, int tag);
void CL_PO_Update(int i, float x, float y, float angle);
void CL_SignonReply();

// PUBLIC FUNCTION PROTOTYPES ----------------------------------------------

// PRIVATE FUNCTION PROTOTYPES ---------------------------------------------

// EXTERNAL DATA DECLARATIONS ----------------------------------------------

// PUBLIC DATA DEFINITIONS -------------------------------------------------

VEntity**		cl_mobjs;
clmobjbase_t*	cl_mo_base;
VEntity*		cl_weapon_mobjs[MAXPLAYERS];
VModel*			model_precache[1024];
VStr			skin_list[256];

// PRIVATE DATA DEFINITIONS ------------------------------------------------

static VModel*			weapon_model_precache[1024];

// CODE --------------------------------------------------------------------

void CL_Clear()
{
	guard(CL_Clear);
	cl->serverinfo.Clean();
	memset((byte*)cl + sizeof(VObject), 0, cl->GetClass()->ClassSize - sizeof(VObject));
	memset(&cl_level, 0, sizeof(cl_level));
	for (int i = 0; i < GMaxEntities; i++)
		if (cl_mobjs[i])
			cl_mobjs[i]->ConditionalDestroy();
	memset(cl_mobjs, 0, sizeof(VEntity*) * GMaxEntities);
	memset(cl_mo_base, 0, sizeof(clmobjbase_t) * GMaxEntities);
	for (int i = 0; i < MAXPLAYERS; i++)
		if (cl_weapon_mobjs[i])
			cl_weapon_mobjs[i]->ConditionalDestroy();
	memset(cl_weapon_mobjs, 0, sizeof(cl_weapon_mobjs));
	memset(cl_dlights, 0, sizeof(cl_dlights));
	for (int i = 0; i < MAXPLAYERS; i++)
	{
		scores[i].name.Clean();
		scores[i].userinfo.Clean();
	}
	memset(scores, 0, sizeof(scores));
	CL_ClearInput();
#ifdef SERVER
	if (!sv.active)
#endif
	{
		// Make sure all sounds are stopped.
		GAudio->StopAllSound();
	}
	for (int i = 0; i < GMaxEntities; i++)
		cl_mobjs[i] = Spawn<VEntity>();
	for (int i = 0; i < MAXPLAYERS; i++)
		cl_weapon_mobjs[i] = Spawn<VEntity>();
	cls.signon = 0;
	unguard;
}

static void CL_ReadMobjBase(VMessage& msg, clmobjbase_t &mobj)
{
	mobj.origin.x = msg.ReadShort();
	mobj.origin.y = msg.ReadShort();
	mobj.origin.z = msg.ReadShort();
	mobj.angles.yaw = ByteToAngle(msg.ReadByte());
	mobj.angles.pitch = ByteToAngle(msg.ReadByte());
	mobj.angles.roll = ByteToAngle(msg.ReadByte());
	mobj.sprite = (word)msg.ReadShort();
	mobj.frame = (word)msg.ReadShort();
	mobj.translucency = msg.ReadByte();
	mobj.translation = msg.ReadByte();
	mobj.effects = msg.ReadByte();
	mobj.model_index = msg.ReadShort();
	mobj.alias_frame = msg.ReadByte();
	mobj.spritetype = mobj.sprite >> 10;
	mobj.sprite &= 0x3ff;
}

static void CL_ParseBaseline(VMessage& msg)
{
	int		i;

	i = msg.ReadShort();

	CL_ReadMobjBase(msg, cl_mo_base[i]);
}

static void CL_ReadMobj(VMessage& msg, int bits, VEntity* mobj, const clmobjbase_t &base)
{
	if (bits & MOB_X)
		mobj->Origin.x = msg.ReadShort();
	else
		mobj->Origin.x = base.origin.x;
	if (bits & MOB_Y)
		mobj->Origin.y = msg.ReadShort();
	else
		mobj->Origin.y = base.origin.y;
	if (bits & MOB_Z)
		mobj->Origin.z = msg.ReadShort();
	else
		mobj->Origin.z = base.origin.z;
	if (bits & MOB_ANGLE)
		mobj->Angles.yaw = ByteToAngle(msg.ReadByte());
	else
		mobj->Angles.yaw = base.angles.yaw;
	if (bits & MOB_ANGLEP)
		mobj->Angles.pitch = ByteToAngle(msg.ReadByte());
	else
		mobj->Angles.pitch = base.angles.pitch;
	if (bits & MOB_ANGLER)
		mobj->Angles.roll = ByteToAngle(msg.ReadByte());
	else
		mobj->Angles.roll = base.angles.roll;
	if (bits & MOB_SPRITE)
	{
		mobj->SpriteIndex = (word)msg.ReadShort();
		mobj->SpriteType = mobj->SpriteIndex >> 10;
		mobj->SpriteIndex &= 0x3ff;
	}
	else
	{
		mobj->SpriteIndex = base.sprite;
		mobj->SpriteType = base.spritetype;
	}
	if (bits & MOB_FRAME)
		mobj->SpriteFrame = (byte)msg.ReadByte();
	else
		mobj->SpriteFrame = base.frame;
	if (bits & MOB_TRANSLUC)
		mobj->Translucency = msg.ReadByte();
	else
		mobj->Translucency = base.translucency;
	if (bits & MOB_TRANSL)
		mobj->Translation = msg.ReadByte();
	else
		mobj->Translation = base.translation;
	if (bits & MOB_EFFECTS)
		mobj->Effects = msg.ReadByte();
	else
		mobj->Effects = base.effects;
	if (bits & MOB_MODEL)
	{
		mobj->ModelIndex = msg.ReadShort();
	}
	else
	{
		mobj->ModelIndex = base.model_index;
	}
	if (bits & MOB_SKIN)
	{
		mobj->ModelSkinIndex = msg.ReadByte();
		if (!mobj->ModelSkinIndex)
		{
			mobj->ModelSkinNum = msg.ReadByte();
		}
		else
		{
			mobj->ModelSkinNum = 0;
		}
	}
	else
	{
		mobj->ModelSkinIndex = 0;
		mobj->ModelSkinNum = 0;
	}
	if (mobj->ModelIndex && (bits & MOB_FRAME))
		mobj->ModelFrame = msg.ReadByte();
	else
		mobj->ModelFrame = base.alias_frame;
}

static void CL_ParseUpdateMobj(VMessage& msg)
{
	guard(CL_ParseUpdateMobj);
	int		i;
	int		bits;

	bits = msg.ReadByte();
	if (bits & MOB_MORE_BITS)
		bits |= msg.ReadByte() << 8;

	if (bits & MOB_BIG_NUM)
		i = msg.ReadShort();
	else
		i = msg.ReadByte();

	CL_ReadMobj(msg, bits, cl_mobjs[i], cl_mo_base[i]);

	//	Marking mobj in use
	cl_mobjs[i]->InUse = 2;

	if (bits & MOB_WEAPON && model_precache[cl_mobjs[i]->ModelIndex] &&
		weapon_model_precache[cl_mobjs[i]->ModelIndex])
	{
		VEntity* ent = cl_mobjs[i];
		VEntity* wpent = cl_weapon_mobjs[i];

		wpent->InUse = true;
		wpent->Origin = ent->Origin;
		wpent->Angles = ent->Angles;
		wpent->ModelIndex = msg.ReadShort();
		wpent->ModelFrame = 1;
		wpent->Translucency = ent->Translucency;

		R_PositionWeaponModel(wpent, weapon_model_precache[ent->ModelIndex],
			ent->ModelFrame);
	}
	else if (bits & MOB_WEAPON)
	{
		msg.ReadShort();
	}
	unguard;
}

static void CL_ParseSecUpdate(VMessage& msg)
{
	int			bits;
	int			i;

	bits = msg.ReadByte();
	if (bits & SUB_BIG_NUM)
		i = msg.ReadShort();
	else
		i = msg.ReadByte();

	if (bits & SUB_FLOOR)
		GClLevel->Sectors[i].floor.dist = msg.ReadShort();
	if (bits & SUB_CEIL)
		GClLevel->Sectors[i].ceiling.dist = msg.ReadShort();
	if (bits & SUB_LIGHT)
		GClLevel->Sectors[i].params.lightlevel = msg.ReadByte() << 2;
	if (bits & SUB_FLOOR_X)
		GClLevel->Sectors[i].floor.xoffs = msg.ReadByte() & 63;
	if (bits & SUB_FLOOR_Y)
		GClLevel->Sectors[i].floor.yoffs = msg.ReadByte() & 63;
	if (bits & SUB_CEIL_X)
		GClLevel->Sectors[i].ceiling.xoffs = msg.ReadByte() & 63;
	if (bits & SUB_CEIL_Y)
		GClLevel->Sectors[i].ceiling.yoffs = msg.ReadByte() & 63;
	if (bits & (SUB_FLOOR | SUB_CEIL))
		CalcSecMinMaxs(&GClLevel->Sectors[i]);
}

static void CL_ParseViewData(VMessage& msg)
{
	int		i;
	int		bits;

	msg >> cl->vieworg.x
		>> cl->vieworg.y
		>> cl->vieworg.z;
	cl->extralight = msg.ReadByte();
	cl->fixedcolormap = msg.ReadByte();
	cl->palette = msg.ReadByte();
	cl->translucency = msg.ReadByte();
	cl->pspriteSY = msg.ReadShort();

	cl->psprites[0].sprite = msg.ReadShort();
	if (cl->psprites[0].sprite != -1)
	{
		cl->psprites[0].frame = msg.ReadByte();
		cl->psprites[0].alias_model = model_precache[msg.ReadShort()];
		cl->psprites[0].alias_frame = msg.ReadByte();
		cl->psprites[0].sx = msg.ReadShort();
		cl->psprites[0].sy = msg.ReadShort();
	}

	cl->psprites[1].sprite = msg.ReadShort();
	if (cl->psprites[1].sprite != -1)
	{
		cl->psprites[1].frame = msg.ReadByte();
		cl->psprites[1].alias_model = model_precache[msg.ReadShort()];
		cl->psprites[1].alias_frame = msg.ReadByte();
		cl->psprites[1].sx = msg.ReadShort();
		cl->psprites[1].sy = msg.ReadShort();
	}

	cl->health = msg.ReadByte();
	msg >> cl->items;
	cl->Frags = msg.ReadShort();

	bits = msg.ReadByte();
	for (i = 0; i < NUM_CSHIFTS; i++)
	{
		if (bits & (1 << i))
			msg >> cl->cshifts[i];
		else
			cl->cshifts[i] = 0;
	}
}

static void CL_ParseStartSound(VMessage& msg)
{
	word		sound_id;
	word		origin_id;
	float		x = 0.0;
	float		y = 0.0;
	float		z = 0.0;
	byte		volume;
	int			channel;

	msg >> sound_id
		>> origin_id;

	channel = origin_id >> 13;
	origin_id &= 0x1fff;

	if (origin_id)
	{
		x = msg.ReadShort();
		y = msg.ReadShort();
		z = msg.ReadShort();
	}
	msg >> volume;

	GAudio->PlaySound(sound_id, TVec(x, y, z), TVec(0, 0, 0), origin_id,
		channel, volume / 127.0);
}

static void CL_ParseStopSound(VMessage& msg)
{
	word	origin_id;
	int		channel;

	msg >> origin_id;

	channel = origin_id >> 13;
	origin_id &= 0x1fff;

	GAudio->StopSound(origin_id, channel);
}

static void CL_ParseStartSeq(VMessage& msg)
{
	int			origin_id;
	float		x;
	float		y;
	float		z;
	const char*	name;

	origin_id = msg.ReadShort();
	x = msg.ReadShort();
	y = msg.ReadShort();
	z = msg.ReadShort();
	name = msg.ReadString();

	GAudio->StartSequenceName(origin_id, TVec(x, y, z), name);
}

static void CL_ParseStopSeq(VMessage& msg)
{
	word	origin_id;

	msg >> origin_id;

	GAudio->StopSequence(origin_id);
}

static void CL_ParseTime(VMessage& msg)
{
	guard(CL_ParseTime);
	float	new_time;
	int		i;

	if (cls.signon == SIGNONS - 1)
	{
		cls.signon = SIGNONS;
		CL_SignonReply();
		GCmdBuf << "HideConsole\n";
	}

	if (cls.signon != SIGNONS)
		Sys_Error("Update when at %d", cls.signon);

	for (i = 0; i < GClLevel->NumSectors; i++)
	{
		sector_t &sec = GClLevel->Sectors[i];
		if (sec.floor.dist != sec.base_floorheight ||
			sec.ceiling.dist != sec.base_ceilingheight)
		{
			sec.floor.dist = sec.base_floorheight;
			sec.ceiling.dist = sec.base_ceilingheight;
			CalcSecMinMaxs(&sec);
		}
		sec.params.lightlevel = sec.base_lightlevel;
		sec.floor.xoffs = 0.0;
		sec.floor.yoffs = 0.0;
		sec.ceiling.xoffs = 0.0;
		sec.ceiling.yoffs = 0.0;
	}

	for (i = 0; i < GClLevel->NumSides; i++)
	{
		GClLevel->Sides[i].textureoffset = GClLevel->Sides[i].base_textureoffset;
		GClLevel->Sides[i].rowoffset = GClLevel->Sides[i].base_rowoffset;
	}

	for (i = 0; i < GMaxEntities; i++)
	{
		if (cl_mobjs[i]->InUse)
		{
			cl_mobjs[i]->InUse--;
		}
	}

	for (i = 0; i < MAXPLAYERS; i++)
	{
		if (cl_weapon_mobjs[i]->InUse)
		{
			cl_weapon_mobjs[i]->InUse--;
		}
	}

	R_AnimateSurfaces();
	msg >> new_time;
	cl_level.tictime = int(new_time * 35);
	cl_level.time = new_time;
	cl->mtime[1] = cl->mtime[0];
	cl->mtime[0] = new_time;
	cl->worldTimer = cl_level.tictime;
	unguard;
}

//==========================================================================
//
//	CL_ReadFromServerInfo
//
//==========================================================================

static void CL_ReadFromServerInfo()
{
	VCvar::SetCheating(!!atoi(*Info_ValueForKey(cl->serverinfo, "sv_cheats")));
}

//==========================================================================
//
//	CL_DoLoadLevel
//
//==========================================================================

void CL_SetupLevel();

static void CL_ParseServerInfo(VMessage& msg)
{
	guard(CL_ParseServerInfo);
	byte		ver;

	msg >> ver;
	if (ver != PROTOCOL_VERSION)
		Host_Error("Server runs protocol %d, not %d", ver, PROTOCOL_VERSION);

	CL_Clear();

	cl->serverinfo = msg.ReadString();
	CL_ReadFromServerInfo();

	cl_level.MapName = msg.ReadString();
	VStr::Cpy(cl_level.level_name, msg.ReadString());

	cl->clientnum = msg.ReadByte();
	cl->maxclients = msg.ReadByte();
	cl->deathmatch = msg.ReadByte();

	msg >> cl_level.totalkills
		>> cl_level.totalitems
		>> cl_level.totalsecret;
	cl_level.sky1Texture = (word)msg.ReadShort();
	cl_level.sky2Texture = (word)msg.ReadShort();
	msg >> cl_level.sky1ScrollDelta
		>> cl_level.sky2ScrollDelta;
	cl_level.doubleSky = msg.ReadByte();
	cl_level.lightning = msg.ReadByte();
	cl_level.SkyBox = msg.ReadString();
	cl_level.FadeTable = msg.ReadString();

	cl_level.SongLump = msg.ReadString();
	cl_level.cdTrack = msg.ReadByte();

	GCon->Log("---------------------------------------");
	GCon->Log(cl_level.level_name);
	GCon->Log("");
	C_ClearNotify();

	CL_LoadLevel(cl_level.MapName);

	//	Temporary hack to restore seen on automap flags.
#ifdef SERVER
	if (sv.active)
	{
		for (int i = 0; i < GClLevel->NumLines; i++)
		{
			GClLevel->Lines[i].flags |= GLevel->Lines[i].flags & ML_MAPPED;
		}
	}
#endif

	R_Start();
	GAudio->Start();

	SB_Start();

	GCon->Log(NAME_Dev, "Client level loaded");
	unguard;
}

//==========================================================================
//
//	CL_ParseIntermission
//
//==========================================================================

static void CL_ParseIntermission(VMessage& msg)
{
	int			i;
	int			j;
	mapInfo_t	ninfo;
	VName		nextmap;

	im.LeaveMap = cl_level.MapName;
	im.LeaveName = cl_level.level_name;
	P_GetMapInfo(cl_level.MapName, ninfo);
	im.leavecluster = ninfo.cluster;

	nextmap = msg.ReadString();
	P_GetMapInfo(nextmap, ninfo);
	im.EnterMap = nextmap;
	im.EnterName = ninfo.name;
	im.entercluster = ninfo.cluster;

	im.totalkills = cl_level.totalkills;
	im.totalitems = cl_level.totalitems;
	im.totalsecret = cl_level.totalsecret;
	im.time = cl_level.time;
	for (i = 0; i < MAXPLAYERS; i++)
	{
		if (msg.ReadByte())
			scores[i].Flags |= scores_t::SF_Active;
		else
			scores[i].Flags &= ~scores_t::SF_Active;
		for (j = 0; j < MAXPLAYERS; j++)
			scores[i].frags[j] = (char)msg.ReadByte();
		scores[i].killcount = msg.ReadShort();
		scores[i].itemcount = msg.ReadShort();
		scores[i].secretcount = msg.ReadShort();
	}
	IM_Start();
}

//==========================================================================
//
//	CL_ParseSpriteList
//
//==========================================================================

static void CL_ParseSpriteList(VMessage& msg)
{
	int count = msg.ReadShort();
	for (int i = 0; i < count; i++)
	{
		R_InstallSprite(msg.ReadString(), i);
	}
}

//==========================================================================
//
//	CL_ParseModel
//
//==========================================================================

static void CL_ParseModel(VMessage& msg)
{
	int i = msg.ReadShort();
	char *name = va("models/%s", msg.ReadString());
	weapon_model_precache[i] = NULL;
	if (FL_FindFile(name))
	{
		model_precache[i] = Mod_FindName(name);
		if (strstr(name, "tris.md2"))
		{
			VStr wpname = VStr(name).ExtractFilePath() + "/weapon.md2";
			if (FL_FindFile(wpname))
			{
				weapon_model_precache[i] = Mod_FindName(*wpname);
			}
			else
			{
				GCon->Logf("Can't find wepon info model %s", *wpname);
			}
		}
	}
	else if (VCvar::GetInt("r_models"))
	{
		GCon->Logf("Can't find %s", name);
	}
}

//==========================================================================
//
//	CL_ParseSkin
//
//==========================================================================

static void CL_ParseSkin(VMessage& msg)
{
	int i = msg.ReadByte();
	skin_list[i] = VStr("models/") + msg.ReadString();
}

//==========================================================================
//
//	CL_ReadFromUserInfo
//
//==========================================================================

static void CL_ReadFromUserInfo(int)
{
}

//==========================================================================
//
//	CL_ParseLineTransuc
//
//==========================================================================

static void CL_ParseLineTransuc(VMessage& msg)
{
	int i = msg.ReadShort();
	int fuzz = msg.ReadByte();
	GClLevel->Lines[i].translucency = fuzz;
}

//==========================================================================
//
//	CL_ParseExtraFloor
//
//==========================================================================

static void CL_ParseExtraFloor(VMessage& msg)
{
	int i = msg.ReadShort();
	int j = msg.ReadShort();
	AddExtraFloor(&GClLevel->Lines[i], &GClLevel->Sectors[j]);
}

//==========================================================================
//
//	CL_ParseHeightSec
//
//==========================================================================

static void CL_ParseHeightSec(VMessage& msg)
{
	sector_t* ToSec = &GClLevel->Sectors[(word)msg.ReadShort()];
	sector_t* HeightSec = &GClLevel->Sectors[(word)msg.ReadShort()];
	int flags = msg.ReadByte();

	ToSec->heightsec = HeightSec;
	if (flags & 2)
	{
		HeightSec->SectorFlags |= sector_t::SF_FakeFloorOnly;
	}
	if (flags & 4)
	{
		HeightSec->SectorFlags |= sector_t::SF_ClipFakePlanes;
	}
	if (flags & 8)
	{
		HeightSec->SectorFlags |= sector_t::SF_UnderWater;
	}
	if (flags & 16)
	{
		HeightSec->SectorFlags |= sector_t::SF_IgnoreHeightSec;
	}
	if (flags & 32)
	{
		HeightSec->SectorFlags |= sector_t::SF_NoFakeLight;
	}
	R_SetupFakeFloors(ToSec);
}

//==========================================================================
//
//	CL_ParseServerMessage
//
//==========================================================================

void CL_ParseServerMessage(VMessage& msg)
{
	guard(CL_ParseServerMessage);
	int			i;
	byte		cmd_type;
	float		x;
	float		y;
	int			tag;
	float		angle;
	VStr		name;
	VStr		string;
	TVec		origin;
	float		radius;
	vuint32		color;
	int			trans;
	sector_t*	sec;

	msg.BeginReading();

	// update command store from the packet
	while (1)
	{
		if (msg.BadRead)
		{
			GCon->Logf(NAME_Dev, "Length %d", msg.CurSize);
			for (i = 0; i < msg.CurSize; i++)
				GCon->Logf(NAME_Dev, "  %d", (int)msg.Data[i]);
			Host_Error("Packet corupted");
		}

		msg >> cmd_type;

		if (msg.BadRead)
			break; // Here this means end of packet

		switch (cmd_type)
		{
		case svc_nop:
			break;

		case svc_disconnect:
			Host_EndGame("Server disconnected");

		case svc_server_info:
			CL_ParseServerInfo(msg);
			break;

		case svc_spawn_baseline:
			CL_ParseBaseline(msg);
			break;

		case svc_update_mobj:
			CL_ParseUpdateMobj(msg);
			break;

		case svc_side_top:
			i = msg.ReadShort();
			GClLevel->Sides[i].toptexture = msg.ReadShort();
			break;

		case svc_side_mid:
			i = msg.ReadShort();
			GClLevel->Sides[i].midtexture = msg.ReadShort();
			break;

		case svc_side_bot:
			i = msg.ReadShort();
			GClLevel->Sides[i].bottomtexture = msg.ReadShort();
			break;

		case svc_side_ofs:
			i = msg.ReadShort();
			GClLevel->Sides[i].textureoffset = msg.ReadShort();
			GClLevel->Sides[i].rowoffset = msg.ReadShort();
			break;

		case svc_sec_floor:
			i = msg.ReadShort();
			GClLevel->Sectors[i].floor.pic = (word)msg.ReadShort();
			break;

		case svc_sec_ceil:
			i = msg.ReadShort();
			GClLevel->Sectors[i].ceiling.pic = (word)msg.ReadShort();
			break;

		case svc_sec_update:
			CL_ParseSecUpdate(msg);
			break;

		case svc_set_angles:
			cl->viewangles.pitch = AngleMod180(ByteToAngle(msg.ReadByte()));
			cl->viewangles.yaw = ByteToAngle(msg.ReadByte());
			cl->viewangles.roll = ByteToAngle(msg.ReadByte());
			break;

		case svc_center_look:
//FIXME
			break;

		case svc_view_data:
			CL_ParseViewData(msg);
			break;

		case svc_start_sound:
			CL_ParseStartSound(msg);
			break;

		case svc_stop_sound:
			CL_ParseStopSound(msg);
			break;

		case svc_start_seq:
			CL_ParseStartSeq(msg);
			break;

		case svc_stop_seq:
			CL_ParseStopSeq(msg);
			break;

		case svc_print:
			C_NotifyMessage(msg.ReadString());
			break;

		case svc_center_print:
			C_CenterMessage(msg.ReadString());
			break;

		case svc_time:
			CL_ParseTime(msg);
			break;

		case svc_poly_spawn:
			x = msg.ReadShort();
			y = msg.ReadShort();
			tag = msg.ReadByte();
			CL_PO_SpawnPolyobj(x, y, tag);
			break;

		case svc_poly_translate:
			x = msg.ReadShort();
			y = msg.ReadShort();
			tag = msg.ReadByte();
			CL_PO_TranslateToStartSpot(x, y, tag);
			break;

		case svc_poly_update:
			i = msg.ReadByte();
			x = msg.ReadShort();
			y = msg.ReadShort();
			angle = ByteToAngle(msg.ReadByte());
			CL_PO_Update(i, x, y, angle);
			break;

		case svc_force_lightning:
			R_ForceLightning();
			break;

		case svc_intermission:
			CL_ParseIntermission(msg);
			break;

		case svc_pause:
			if (msg.ReadByte())
			{
				cl->ClientFlags |= VClientState::CF_Paused;
				GAudio->PauseSound();
			}
			else
			{
				cl->ClientFlags &= ~VClientState::CF_Paused;
				GAudio->ResumeSound();
			}
			break;

		case svc_stats_long:
			i = msg.ReadByte();
			msg >> ((int*)((byte*)cl + sizeof(VClientState)))[i];
			break;

		case svc_stats_short:
			i = msg.ReadByte();
			((int*)((byte*)cl + sizeof(VClientState)))[i] = msg.ReadShort();
			break;

		case svc_stats_byte:
			i = msg.ReadByte();
			((int*)((byte*)cl + sizeof(VClientState)))[i] = msg.ReadByte();
			break;

		case svc_stringcmd:
			GCmdBuf << msg.ReadString();
			break;

		case svc_signonnum:
			i = msg.ReadByte();
			if (i <= cls.signon)
				Host_Error("Received signon %i when at %i", i, cls.signon);
			cls.signon = i;
			CL_SignonReply();
			break;

		case svc_skip_intermission:
			IM_SkipIntermission();
			break;

		case svc_finale:
			F_StartFinale();
			break;

		case svc_sec_floor_plane:
			i = msg.ReadShort();
			msg >> GClLevel->Sectors[i].floor.normal.x
				>> GClLevel->Sectors[i].floor.normal.y
				>> GClLevel->Sectors[i].floor.normal.z
				>> GClLevel->Sectors[i].floor.dist;
			GClLevel->Sectors[i].base_floorheight = GClLevel->Sectors[i].floor.dist;
			CalcSecMinMaxs(&GClLevel->Sectors[i]);
			break;

		case svc_sec_ceil_plane:
			i = msg.ReadShort();
			msg >> GClLevel->Sectors[i].ceiling.normal.x
				>> GClLevel->Sectors[i].ceiling.normal.y
				>> GClLevel->Sectors[i].ceiling.normal.z
				>> GClLevel->Sectors[i].ceiling.dist;
			GClLevel->Sectors[i].base_ceilingheight  = GClLevel->Sectors[i].ceiling.dist;
			CalcSecMinMaxs(&GClLevel->Sectors[i]);
			break;

		case svc_serverinfo:
			name = msg.ReadString();
			string = msg.ReadString();
			Info_SetValueForKey(cl->serverinfo, name, string);
			CL_ReadFromServerInfo();
			break;

		case svc_userinfo:
			i = msg.ReadByte();
			scores[i].userinfo = msg.ReadString();
			CL_ReadFromUserInfo(i);
			break;

		case svc_setinfo:
			i = msg.ReadByte();
			name = msg.ReadString();
			string = msg.ReadString();
			Info_SetValueForKey(scores[i].userinfo, name, string);
			CL_ReadFromUserInfo(i);
			break;

		case svc_sprites:
			CL_ParseSpriteList(msg);
			break;

		case svc_model:
			CL_ParseModel(msg);
			break;

		case svc_skin:
			CL_ParseSkin(msg);
			break;

		case svc_line_transluc:
			CL_ParseLineTransuc(msg);
			break;

		case svc_sec_transluc:
			i = msg.ReadShort();
			trans = msg.ReadByte();
			sec = &GClLevel->Sectors[i];
			sec->floor.translucency = trans;
			sec->ceiling.translucency = trans;
			for (i = 0; i < sec->linecount; i++)
			{
				sec->lines[i]->translucency = trans;
			}
			break;

		case svc_extra_floor:
			CL_ParseExtraFloor(msg);
			break;

		case svc_swap_planes:
			i = msg.ReadShort();
			SwapPlanes(&GClLevel->Sectors[i]);
			break;

		case svc_static_light:
			origin.x = msg.ReadShort();
			origin.y = msg.ReadShort();
			origin.z = msg.ReadShort();
			radius = (byte)msg.ReadByte() * 8;
			R_AddStaticLight(origin, radius, 0xffffffff);
			break;

		case svc_static_light_rgb:
			origin.x = msg.ReadShort();
			origin.y = msg.ReadShort();
			origin.z = msg.ReadShort();
			radius = (byte)msg.ReadByte() * 8;
			msg >> color;
			R_AddStaticLight(origin, radius, color);
			break;

		case svc_sec_light_color:
			sec = &GClLevel->Sectors[msg.ReadShort()];
			sec->params.LightColor = (msg.ReadByte() << 16) |
				(msg.ReadByte() << 8) | msg.ReadByte();
			break;

		case svc_change_sky:
			cl_level.sky1Texture = (word)msg.ReadShort();
			cl_level.sky2Texture = (word)msg.ReadShort();
			R_SkyChanged();
			break;

		case svc_change_music:
			cl_level.SongLump = msg.ReadString();
			cl_level.cdTrack = msg.ReadByte();
			GAudio->MusicChanged();
			break;

		case svc_set_floor_light_sec:
			i = (word)msg.ReadShort();
			GClLevel->Sectors[i].floor.LightSourceSector = msg.ReadShort();
			break;

		case svc_set_ceil_light_sec:
			i = (word)msg.ReadShort();
			GClLevel->Sectors[i].ceiling.LightSourceSector = msg.ReadShort();
			break;

		case svc_set_heightsec:
			CL_ParseHeightSec(msg);
			break;

		default:
			if (GClGame->eventParseServerCommand(cmd_type))
			{
				break;
			}
			GCon->Logf(NAME_Dev, "Length %d", msg.CurSize);
			for (i = 0; i < msg.CurSize; i++)
			{
				GCon->Logf(NAME_Dev, "  %d", (int)msg.Data[i]);
			}
			GCon->Logf(NAME_Dev, "ReadCount %d", msg.ReadCount);
			Host_Error("Invalid packet %d", cmd_type);
			break;
		}
	}
	unguard;
}
